#include "fw/dynmx/protocol.h"

#include <emlabcpp/algorithm.h>
#include <emlabcpp/iterators/convert.h>
#include <emlabcpp/static_circular_buffer.h>

#pragma once

namespace schpin
{

struct dynmx_read_request
{
        std::size_t count;
};

template < typename Message >
class dynmx_sequencer
{
        // HEADER: 0xFF,0xFF,0xFD,0x00,<id>,<len_low>,<len_high>,<ins>

        enum states
        {
                IDLE,
                HEADER,
                PACKET
        };

public:
        using message_type                                     = Message;
        static constexpr std::array< uint8_t, 4 > header_start = { 0xFF, 0xFF, 0xFD, 0x00 };
        static constexpr std::size_t              header_size  = 7;
        static constexpr std::size_t max_size = em::max( Message::max_size, 2 * header_size );

private:
        states                                          state_;
        em::static_circular_buffer< uint8_t, max_size > buffer_;

public:
        dynmx_read_request start_read_packet()
        {
                state_ = HEADER;
                return { header_size };
        }

        int find_header_start()
        {
                for ( int i = 0; i < static_cast< int >( buffer_.size() - header_start.size() );
                      i++ ) {
                        auto to_check = em::view_n( buffer_.begin(), header_start.size() );
                        if ( em::equal( to_check, header_start ) ) {
                                return i;
                        }
                }
                return -1;
        }

        void truncate_buffer()
        {
                while ( buffer_.size() > header_size ) {
                        buffer_.pop_front();
                }
        }

        std::size_t get_remaining_size()
        {
                auto val = static_cast< std::size_t >( ( buffer_[6] << 8 ) + buffer_[5] );
                return header_size + val - buffer_.size();
        }

        template < typename Iterator >
        em::either< dynmx_read_request, Message > __attribute__( ( noinline ) )
        isr_read_callback( em::view< Iterator > data )
        {
                std::copy( data.begin(), data.end(), std::back_inserter( buffer_ ) );
                int i = 0;
                switch ( state_ ) {
                        case IDLE:
                                break;
                        case HEADER:
                                i = find_header_start();
                                if ( i == -1 ) {
                                        truncate_buffer();
                                        return dynmx_read_request{ 4 };
                                }
                                state_ = PACKET;
                                for ( ; i > 0; i-- ) {
                                        buffer_.pop_front();
                                }
                                return dynmx_read_request{ get_remaining_size() };
                        case PACKET:
                                if ( get_remaining_size() == 0 ) {
                                        auto opt_msg = Message::make( buffer_ );
                                        return *opt_msg;
                                }
                                return dynmx_read_request{ get_remaining_size() };
                }
                return dynmx_read_request{ 0 };
        }
};
}  // namespace schpin
